@@ -117,16 +117,12 @@ color_combo_get_type (void)
  * Fires signal "changed" with the current color as its param
  */
 static void
-emit_change (ColorCombo *cc)
+emit_change (ColorCombo *cc, GdkColor *color)
 {
-	GdkColor *color;
-
-	color = color_palette_get_current_color (cc->palette);
   	gtk_signal_emit (
 		GTK_OBJECT (cc),
 		color_combo_signals [CHANGED],
 		color);
-	g_free (color);
 	gtk_combo_box_popup_hide (GTK_COMBO_BOX (cc));
 }
 
@@ -134,15 +130,15 @@ static void
 cb_color_change (ColorPalette *P, GdkColor *color, gboolean custom, ColorCombo *cc)
 {
 	set_color(cc, color);
-	emit_change(cc);
+	emit_change (cc, color);
 }
 
-
-
 static void
 preview_clicked (GtkWidget *button, ColorCombo *cc)
 {
-	emit_change (cc);
+	GdkColor *color = color_palette_get_current_color (cc->palette);
+	emit_change (cc, color);
+	g_free (color);
 }
 
 static void
@@ -205,9 +201,9 @@ color_combo_construct (ColorCombo *cc, char **icon,
 	gtk_widget_pop_colormap ();
 	gtk_widget_pop_visual ();
 
-	image = gdk_imlib_create_image_from_xpm_data (icon);
-
 	gnome_canvas_set_scroll_region (cc->preview_canvas, 0, 0, 24, 24);
+	if (icon) {
+		image = gdk_imlib_create_image_from_xpm_data (icon);
 
 	gnome_canvas_item_new (
 		GNOME_CANVAS_GROUP (gnome_canvas_root (cc->preview_canvas)),
@@ -230,6 +226,18 @@ color_combo_construct (ColorCombo *cc, char **icon,
 		"fill_color", "black",
 		"width_pixels", 1,
 		NULL);
+	} else
+		cc->preview_color_item = gnome_canvas_item_new (
+								GNOME_CANVAS_GROUP (gnome_canvas_root (cc->preview_canvas)),
+								gnome_canvas_rect_get_type (),
+								"x1",         2.0,
+								"y1",         1.0,
+								"x2",         21.0,
+								"y2",         22.0,
+								"fill_color", "black",
+								"width_pixels", 1,
+								NULL);
+
 	gtk_container_add (GTK_CONTAINER (cc->preview_button), GTK_WIDGET (cc->preview_canvas));
 	gtk_widget_set_usize (GTK_WIDGET (cc->preview_canvas), 24, 24);
 	gtk_signal_connect (GTK_OBJECT (cc->preview_button), "clicked",
@@ -263,8 +271,6 @@ color_combo_new (char **icon, char const * const no_color_label,
 {
 	ColorCombo *cc;
 
-	g_return_val_if_fail (icon != NULL, NULL);
-
 	cc = gtk_type_new (color_combo_get_type ());
 
         cc->default_color = default_color;
